@@ -1,21 +1,31 @@
 #!/usr/bin/env python3
 
+import sys
 from time import sleep
-from akinaka.libs import helpers
 from akinaka.libs import exceptions
-from botocore.exceptions import ParamValidationError
-from botocore.exceptions import ClientError
 from akinaka.client.aws_client import AWS_Client
+import botocore.exceptions
 import logging
+from pprint import pformat
 
 aws_client = AWS_Client()
 
-class ASG():
+
+def log_new_asg_name(new_asg_name):
+    """ Write [new_asg_name] to 'inactive_asg.txt' """
+
+    logging.info("ASG fully healthy. Logging new ASG name to \"inactive_asg.txt\"")
+    with open("inactive_asg.txt", "w", encoding="utf8") as this_file:
+        this_file.write(new_asg_name)
+
+
+class ASG(): # pylint: disable=too-many-public-methods
     """All the methods needed to perform a blue/green deploy"""
 
-    def __init__(self, region, role_arn):
+    def __init__(self, region, role_arn, log_level):
         self.region = region
         self.role_arn = role_arn
+        logging.getLogger().setLevel(log_level)
 
     def get_application_name(self, asg, loadbalancer=None, target_group=None):
         """
@@ -47,12 +57,19 @@ class ASG():
 
         return {"inactive_asg": inactive_asg, "active_asg": active_asg}
 
-    def scale_down_inactive(self, asg):
-        # NOTE: We're making the heavy presumption here that the _active_ ASG is the one we've been
-        #       given from the command line
-        inactive_asg = self.get_inactive_asg(asg)
+    def scale_down_inactive(self, target_group):
+        """
+        Given [target_group], work out which ASG is the inactive one, and scale it down to 0.
+
+        Returns True on success, or calls sys.exit(1) on failure (via self.scale())
+        """
+        target_group_arn = self.get_target_group_arn(target_group=target_group)
+        active_asg = self.get_active_asg(target_group_arn)
+        inactive_asg = self.get_inactive_asg(active_asg)
         self.scale(inactive_asg, 0, 0, 0)
 
+        return True
+
     def set_asg_launch_template_version(self, asg, lt_id, lt_version):
         asg_client = aws_client.create_client('autoscaling', self.region, self.role_arn)
 
@@ -64,88 +81,162 @@ class ASG():
             }
         )
 
-    def do_update(self, ami, asg=None, loadbalancer=None, target_group=None):
-        asg_liveness_info = self.asgs_by_liveness(asg=asg, loadbalancer=loadbalancer, target_group=target_group)
+    def rescale(self, active_asg, inactive_asg):
+        """
+        Scales [inactive_asg] to the same values as those found in [active_asg]
 
-        inactive_asg = asg_liveness_info['inactive_asg']
-        active_asg = asg_liveness_info['active_asg']
-        new_ami = ami
+        Returns True on success and exits with a code of 1 if scaling failed
+        """
 
-        logging.info("New ASG was worked out as {}. Now updating it's Launch Template".format(inactive_asg))
+        active_asg_size = self.get_current_scale(active_asg)
 
-        # Update the lt and set the soon to be new ASG to the new launch template version
-        updated_lt = self.update_launch_template(new_ami, self.get_lt_name(inactive_asg))
-        self.set_asg_launch_template_version(
-            asg=inactive_asg,
-            lt_id=updated_lt["id"],
-            lt_version=updated_lt["version"]
-        )
-
-        scale_to = self.get_current_scale(active_asg)
-
-        logging.info("Scaling ASG '{}' down to ({}, {}, {})".format(inactive_asg, 0, 0, 0))
+        logging.info('Scaling to 0 first to start with a clean slate')
         self.scale(inactive_asg, 0, 0, 0)
-        while not self.asg_is_empty(inactive_asg):
-            logging.info("Waiting for instances in ASG '{}' to terminate".format(inactive_asg))
-            sleep(10)
-
-        logging.info("Scaling ASG '{}' back up to ({}, {}, {})".format(inactive_asg, scale_to['min'], scale_to['max'], scale_to['desired']))
+        logging.info(f"Scaling ASG {inactive_asg} to {active_asg_size['min']}, {active_asg_size['max']}, {active_asg_size['desired']}")
+        logging.info('Scaling to match the numbers from the active ASG')
         self.scale(
-            auto_scaling_group_id = inactive_asg,
-            min_size = scale_to['min'],
-            max_size = scale_to['max'],
-            desired =  scale_to['desired']
+            asg = inactive_asg,
+            min_size = active_asg_size['min'],
+            max_size = active_asg_size['max'],
+            desired =  active_asg_size['desired']
         )
 
-        while len(self.get_auto_scaling_group_instances(inactive_asg)) < 1:
-            logging.info("Waiting for instances in ASG to start ...")
-            sleep(10)
+        return True
 
-        logging.info("First instance has started")
+    def asg_instance_list(self, asg):
+        """
+        TODO: Can this replace get_auto_scaling_group_instances()?
 
-        # Try to get information for an instance in the new ASG 20 times
-        for i in range(20):
-            try:
-                first_new_instance = self.get_first_new_instance(inactive_asg)
-                if i == 20:
-                    raise exceptions.AkinakaCriticalException
-                break
-            except exceptions.AkinakaLoggingError as error:
-                logging.info("Retry {}".format(i))
-                logging.info("Problem in getting data for first new ASG instance")
-                logging.info("get_auto_scaling_group_instances() returned: {}".format(self.get_auto_scaling_group_instances(inactive_asg)))
-                logging.info("Error was: {}".format(error))
+        Return a list of instances from [asg]
+        """
 
-        # Show console output for first instance up until it's Lifecycle Hook has passed
-        while self.get_auto_scaling_group_instances(auto_scaling_group_id=inactive_asg, instance_ids=[first_new_instance])[0]['LifecycleState'] != "InService":
-            try:
-                logging.info("Attempting to retrieve console output from first instance up -- this will not work for non-nitro hypervisor VMs")
-                logging.info(self.get_instance_console_output(first_new_instance)['Output'])
-            except KeyError:
-                logging.info("No output from instance yet. Trying again in 10 seconds.")
-            sleep(10)
+        asg_client = aws_client.create_client('autoscaling', self.region, self.role_arn)
+        asg_info = asg_client.describe_auto_scaling_groups(AutoScalingGroupNames=[asg])
+        return asg_info['AutoScalingGroups'][0]['Instances']
 
-        # Gets the actual configured sizes of the inactive ASG
-        inactive_asg_size = self.get_current_scale(inactive_asg)
-        logging.info("'{}' sizes currently set to (min: {}, max: {}, desired: {})".format(inactive_asg, inactive_asg_size['min'], inactive_asg_size['max'], inactive_asg_size['desired']))
+    def last_instance_loglines(self, instances: list, sleep_duration=None):
+        """ Returns a dict of { instance_id: log } containing the last log line for [instances] """
 
-        # Wait for remaining instances (if any) to come up too (up to 10 minutes = 300 attempts + 1 sec sleep on each attempt)
+        sleep_duration = sleep_duration or 0
+        logging.info(f"Sleeping for {sleep_duration}")
+        sleep(sleep_duration)
+        instance_loglines = {}
+        logs_client = aws_client.create_client('logs', self.region, self.role_arn)
+
+        try:
+            for this_instance in instances:
+                event_stream = logs_client.get_log_events(logGroupName='instance', logStreamName=f"cloud-init-output-{this_instance}")
+                last_log_line = [ this_event['message'] for this_event in event_stream['events'] ][-1]
+                instance_loglines[this_instance] = last_log_line
+        except botocore.exceptions.ClientError as error:
+            logging.error(f"{error} When trying to fetch the stream cloud-init-output-{this_instance} from the log group called 'instance'")
+
+        return instance_loglines
+
+    def scale_waiter(self, asg, desired_scale, timeout=600):
+        """
+        Scales [asg] to [desired_scale] and waits [timeout] seconds for all instances
+        to become healthy. [timeout] defaults to 600
+
+        Returns True on success, or False on failure
+        """
+
+        max_attempts = timeout/20
         attempts = 0
-        max_attempts = 300
-        while len(self.asgs_healthy_instances(inactive_asg)) < inactive_asg_size['desired'] and attempts != max_attempts:
-            logging.info("Waiting for all instances to be healthy ...")
-            sleep(1)
+
+        while len(self.asg_instance_list(asg)) != desired_scale or (len(self.asgs_instances_by_lifecycle(asg)["in_service"]) < desired_scale and attempts != max_attempts):
+            logging.info("Waiting for scaling event to finish successfully. Next poll in 20 seconds")
+
+            sleep(20)
             attempts += 1
             if attempts == max_attempts:
-                logging.info("Max timeout reached without success... Exiting!")
-                raise exceptions.AkinakaLoggingError
+                out_of_service_info = self.asgs_instances_by_lifecycle(asg)["not_in_service"]
+                out_of_service_instances = [ this_instance['InstanceId'] for this_instance in out_of_service_info ]
+                last_instance_loglines = self.last_instance_loglines(out_of_service_instances, 5)
 
-        logging.info("ASG fully healthy. Logging new ASG name to \"inactive_asg.txt\"")
-        open("inactive_asg.txt", "w").write(inactive_asg)
+                logging.info("Timeout reached without success whilst waiting for all instances to become healthy")
+                logging.info(f"""
+The following instances were never promoted to 'InService':
+
+{out_of_service_instances}
+
+Their last loglines (sleeping for 5 seconds first to give CloudWatch time to get them):
+
+{pformat(last_instance_loglines)}
+                """)
+                return False
+
+        return True
+
+    def wait_for_clean_asg_refresh_status(self, asg, acceptable_statuses, timeout, fail_on_failure=False):
+        """
+        Go into a loop for a maximum of [timeout] seconds, waiting for the status
+        of the ASG refresh for [asg] to one of [acceptable_statuses]
+
+        Return {success: True, status: current_status} when one of [acceptable_statuses] was
+        received, and {success: False, status: current_status} when the [timeout] was reached,
+        or if [fail_on_failure=True] and receive a failure status. Default for [fail_on_failure]
+        is False, because the failure codes may be within [acceptable_statuses]
+        """
+
+        asg_client = aws_client.create_client('autoscaling', self.region, self.role_arn)
+
+        time_count = 0
+        while True:
+            try:
+                result = asg_client.describe_instance_refreshes(AutoScalingGroupName=asg)['InstanceRefreshes'][0]
+            except IndexError:
+                return {'success': True, 'status': 'NoPreviousDeploy'}
+
+            current_status = result['Status']
+
+            if current_status in acceptable_statuses:
+                return {'success': True, 'status': current_status}
+
+            if time_count >= timeout:
+                logging.info("Timeout reached")
+                return {'success': False, 'status': current_status}
+
+            if fail_on_failure and (current_status == 'Failed' or current_status == 'Cancelling'):
+                return {'success': False, 'status': current_status}
+
+            logging.info(f"\nCurrent Status: {result}\nTime counter: {time_count}")
+
+            time_count += 20
+            sleep(20)
+
+    def refresh_asg(self, asg):
+        """
+        Triggers and monitors an ASG refresh call of [asg]
+
+        Returns True on success and False on failure
+        """
+
+        if not self.wait_for_clean_asg_refresh_status(asg, ['Successful', 'Cancelled', 'Failed'], 60)['success']:
+            logging.error("Timeout reached whilst waiting for ASG to become ready for another event. Try again later.")
+            return False
+
+        logging.info(f"Starting instance refresh for ASG {asg}")
+        asg_client = aws_client.create_client('autoscaling', self.region, self.role_arn)
+        asg_client.start_instance_refresh(
+            AutoScalingGroupName=asg,
+            Preferences={
+                'MinHealthyPercentage': 100
+            }
+        )
+
+        cooldown_period = asg_client.describe_auto_scaling_groups(AutoScalingGroupNames=[asg])['AutoScalingGroups'][0]['DefaultCooldown']
+        timeout = cooldown_period + 300
+        if not self.wait_for_clean_asg_refresh_status(asg, ['Successful'], timeout, fail_on_failure=True)['success']:
+            logging.error("The rollout failed")
+            return False
+
+        return True
 
     def get_first_new_instance(self, new_asg):
         asg_instances = self.get_auto_scaling_group_instances(new_asg)
-        if len(asg_instances) < 1: raise exceptions.AkinakaLoggingError
+        if len(asg_instances) < 1:
+            raise exceptions.AkinakaLoggingError
 
         return asg_instances[0]['InstanceId']
 
@@ -180,10 +271,16 @@ class ASG():
 
         Both are mutually exclusive, and at least one must be supplied
         """
-
         alb_client = aws_client.create_client('elbv2', self.region, self.role_arn)
         if target_group is not None:
-            return alb_client.describe_target_groups(Names=[target_group])['TargetGroups'][0]['TargetGroupArn']
+            try:
+                return alb_client.describe_target_groups(Names=[target_group])['TargetGroups'][0]['TargetGroupArn']
+            except alb_client.exceptions.TargetGroupNotFoundException as e:
+                logging.error(f"Couldn't describe the target group {target_group}. Does it exist?\n{e}")
+                all_target_groups_info = alb_client.describe_target_groups()['TargetGroups']
+                all_target_groups = [ tg['TargetGroupName'] for tg in all_target_groups_info ]
+                logging.error(f"Here's a list of all target groups I could find. Some may belong to different deployments: {all_target_groups}")
+                sys.exit(1)
 
         loadbalancer_raw_info = alb_client.describe_load_balancers(Names=[loadbalancer])
         loadbalancer_arn = loadbalancer_raw_info['LoadBalancers'][0]['LoadBalancerArn']
@@ -313,24 +410,30 @@ class ASG():
             "version": launch_template_new_version
         }
 
-    def scale(self, auto_scaling_group_id, min_size, max_size, desired):
-        """Scale an ASG to {'min_size', 'max_size', 'desired'}"""
+    def scale(self, asg, min_size, max_size, desired):
+        """
+        Scale [asg] to {'min_size', 'max_size', 'desired'}, using scale_waiter()
+        to wait for the ASG to be healthy again
+
+        Returns the response from the call on success, or calls sys.exit(1) on failure
+        """
 
         asg_client = aws_client.create_client('autoscaling', self.region, self.role_arn)
-
         response = asg_client.update_auto_scaling_group(
-            AutoScalingGroupName=auto_scaling_group_id,
+            AutoScalingGroupName=asg,
             MinSize=min_size,
             MaxSize=max_size,
             DesiredCapacity=desired
         )
 
+        if self.scale_waiter(asg, desired) == False:
+            sys.exit(1)
+
         return response
 
     def get_current_scale(self, asg):
         """
-        Returns the current scales of the given ASG as dict {'desired', 'min', 'max'},
-        expects ASG ID as argument
+        Returns the current scales of [asg] as dict {'desired', 'min', 'max'}
         """
 
         asg_client = aws_client.create_client('autoscaling', self.region, self.role_arn)
@@ -352,7 +455,7 @@ class ASG():
             if i['AutoScalingGroupName'] == auto_scaling_group_id:
                 target_instances.append(i)
 
-                logging.debug("Instance {instance_id} has state = {instance_state},"
+                logging.info("Instance {instance_id} has state = {instance_state}, "
                     "Lifecycle is at {instance_lifecycle_state}".format(
                         instance_id = i['InstanceId'],
                         instance_state = i['HealthStatus'],
@@ -374,12 +477,46 @@ class ASG():
 
         return all(instance.get('LifecycleState') == 'Terminated' for instance in asg_instances)
 
-    def asgs_healthy_instances(self, auto_scaling_group_id):
+    def asgs_instances_by_lifecycle(self, auto_scaling_group_id):
+        """ Returns a dict of { (list)in_service, (list)not_in_service } """
+
         asg_instances = self.get_auto_scaling_group_instances(auto_scaling_group_id)
-        healthy_instances = []
+        instances_by_lifecycle = { "in_service": [], "not_in_service": [] }
 
         for i in asg_instances:
             if i['LifecycleState'] == "InService":
-                healthy_instances.append(i)
+                instances_by_lifecycle["in_service"].append(i)
+            else:
+                instances_by_lifecycle["not_in_service"].append(i)
 
-        return healthy_instances
+        return instances_by_lifecycle
+
+    def main(self, ami, asg=None, loadbalancer=None, target_group=None):
+        """
+        Calls necessary methods to perform an update:
+
+        1. Figures out which ASGs are active and inactive
+        2. Creates new launch template version with AMI set to [ami]
+        3. Scales inactive ASG down, then back up using the new launch template version
+        """
+
+        asg_liveness_info = self.asgs_by_liveness(asg=asg, loadbalancer=loadbalancer, target_group=target_group)
+
+        inactive_asg = asg_liveness_info['inactive_asg']
+        active_asg = asg_liveness_info['active_asg']
+        new_ami = ami
+
+        logging.info("New ASG was worked out as {}. Now updating it's Launch Template".format(inactive_asg))
+
+        updated_lt = self.update_launch_template(new_ami, self.get_lt_name(inactive_asg))
+        self.set_asg_launch_template_version(
+            asg=inactive_asg,
+            lt_id=updated_lt["id"],
+            lt_version=updated_lt["version"]
+        )
+
+        if active_asg == inactive_asg:
+            self.refresh_asg(active_asg)
+        else:
+            self.rescale(active_asg, inactive_asg)
+            log_new_asg_name(inactive_asg)
